#define pr_fmt(fmt) KBUILD_MODNAME ": " fmt

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/ftrace.h>
#include <linux/binfmts.h>
#include <linux/kobject.h>
#include <linux/sysfs.h>
#include <linux/string.h>
#include <linux/kallsyms.h>
#include <linux/mman.h>
#include <linux/slab.h>
#include <linux/security.h>
#include <linux/mount.h>
#include <net/sock.h>
#include <linux/netlink.h>
#include <linux/skbuff.h>


struct exec_path {
        struct list_head list;
        char *path;
};

static LIST_HEAD(exec_path_list);
static DEFINE_RWLOCK(exec_path_list_lock);
static struct kobject *exec_kobject = NULL;

struct sock *nl_sock = NULL;
#define MYPROTO NETLINK_USERSOCK
#define MYGRP 17

int app_control_enable = 0;

static void app_send_broadcast_skb(const char *appname)
{
    struct sk_buff *skb;
    struct nlmsghdr *nlh;
    int msg_size = strlen(appname) + 1;
    int res;

    skb = nlmsg_new(NLMSG_ALIGN(msg_size + 1), GFP_KERNEL);
    if (!skb) {
        return;
    }

    nlh = nlmsg_put(skb, 0, 1, NLMSG_DONE, msg_size + 1, 0);
    strcpy(nlmsg_data(nlh), appname);
    NETLINK_CB(skb).dst_group = MYGRP;

    pr_info("Sending skb.\n");
    res = nlmsg_multicast(nl_sock, skb, 0, MYGRP, GFP_KERNEL);
}

struct netlink_kernel_cfg cfg = {
};

void empty_exec_path(void)
{
	write_lock(&exec_path_list_lock);
	while (!list_empty(&exec_path_list)) {
		struct exec_path* path;
		path = list_first_entry(&exec_path_list, struct exec_path, list);
		list_del(&path->list);
		kfree(path);
	}
	write_unlock(&exec_path_list_lock);
}

int load_exec_path(char *buf)
{
	int len = 0;
	char *p1, *p2;
	struct exec_path *path = NULL;
	
	if(!buf)
		return 0;
	
	p1 = p2 = buf;
	while((*p2 == ' ') || (*p2 == '\t'))
		p2++;
	p1 = p2;
	if (*p2 == '\0' || *p2 == '\n')
		return 0;
	
	len = strlen(buf);
	if (*(p2 + len - 1) == '\n')
		*(p2 + len - 1) = '\0';

	while(1) {
		while((*p2 != ' ') && (*p2 != '\t') && (*p2 != '\0'))
			p2++;
		
		if (p1 != p2) {
			path = kmalloc(sizeof(*path) + (p2 - p1 + 1), GFP_ATOMIC);
			if (!path)
				return -ENOMEM;
			
			path->path = (char *)path + sizeof(*path);
			strncpy(path->path, p1, p2 - p1);
			path->path[p2-p1] = '\0';
			write_lock(&exec_path_list_lock);
			list_add_tail(&path->list, &exec_path_list);
			write_unlock(&exec_path_list_lock);
		}
		
		if (*p2 == '\0')
			return 0;
		
		while((*p2 == ' ') || (*p2 == '\t'))
		p2++;
		p1 = p2;
	}
}


ssize_t exec_path_show(struct kobject *kobj, struct kobj_attribute *attr,
                      char *buf)
{
	char *p = buf;
	int len, total = 0;
	struct exec_path *path;
	
	read_lock(&exec_path_list_lock);
	list_for_each_entry(path, &exec_path_list, list) {
		sprintf(p, "%s\n", path->path);
		len = strlen(path->path) + 1;
		p += len;
		total += len;
	}
	read_unlock(&exec_path_list_lock);
	
	return total;
}


static ssize_t exec_path_store(struct kobject *kobj, struct kobj_attribute *attr,
                      const char *buf, size_t count)
{
	int ret = 0;
	if(count == 0)
		return -EFAULT;
	
	empty_exec_path();
	ret = load_exec_path((char *)buf);
	return count;

}

ssize_t enable_show(struct kobject *kobj, struct kobj_attribute *attr,
                      char *buf)
{
	return sprintf(buf, "%d\n", app_control_enable);
}

static ssize_t enable_store(struct kobject *kobj, struct kobj_attribute *attr,
                      const char *buf, size_t count)
{
	int enable = 0;

	if (!strncmp(buf, "1", 1))
		app_control_enable = 1;
	else if (!strncmp(buf, "0", 1))
		app_control_enable = 0;
	else
		return -EFAULT;
	return count;
}

static struct kobj_attribute exec_path_attribute =__ATTR(pathes, 0644, exec_path_show,
                                                   exec_path_store);
static struct kobj_attribute enable_attribute =__ATTR(enable, 0644, enable_show, enable_store);


int file_skip_exec_check(const char *file)
{
	int rc = -1;
	struct exec_path *path;

	if (!app_control_enable)
		return rc;

	if (!file) {
		return rc;
	} else {
		read_lock(&exec_path_list_lock);
		list_for_each_entry(path, &exec_path_list, list) {
			if (strlen(file) < strlen(path->path))
				continue;
			if (!strncmp(file, path->path, strlen(path->path))) {
				app_send_broadcast_skb(file);
				printk(KERN_ERR "Blocked by app control framework: %s\n",
					       		kbasename(file));
				rc = 0;
				break;
			}
		}
		read_unlock(&exec_path_list_lock);
	}
	return rc;
}
EXPORT_SYMBOL(file_skip_exec_check);

static int exec_path_init(void)
{
	int ret;

	exec_kobject = kobject_create_and_add("exec_path", kernel_kobj);
	if (!exec_kobject)
		return -ENOMEM;

	ret = sysfs_create_file(exec_kobject, &exec_path_attribute.attr);
	if (ret)
		goto end1;

	ret = sysfs_create_file(exec_kobject, &enable_attribute.attr);
	if (ret)
		goto end2;

	nl_sock = netlink_kernel_create(&init_net, MYPROTO, &cfg);
	if (!nl_sock)
		goto end3;

	return 0;

end3:
	sysfs_remove_file(exec_kobject, &enable_attribute.attr);
end2:
	sysfs_remove_file(exec_kobject, &exec_path_attribute.attr);
end1:
	kobject_del(exec_kobject);
	empty_exec_path();
	exec_kobject = NULL;

	return ret;
}

static void exec_path_exit(void)
{
	netlink_kernel_release(nl_sock);
	sysfs_remove_file(exec_kobject, &enable_attribute.attr);
	sysfs_remove_file(exec_kobject, &exec_path_attribute.attr);
	kobject_del(exec_kobject);
	empty_exec_path();
	exec_kobject = NULL;
}

module_init(exec_path_init);
module_exit(exec_path_exit);
MODULE_LICENSE("GPL");

