/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright (c) 2022 HYGON Corporation . All rights reserved.
 * Author: Ge Yang <yangge@hygon.cn>
 */

#include <linux/version.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/miscdevice.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/slab.h>
#include <linux/cdev.h>
#include <linux/sched.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 13, 0)
#include <linux/sched/signal.h>
#else
#include <linux/signal.h>
#endif
#include <linux/wait.h>
#include <linux/uuid.h>
#include <linux/vfio.h>
#include <linux/iommu.h>
#include <linux/sysfs.h>
#include <linux/ctype.h>
#include <linux/file.h>
#include <linux/pagemap.h>
#include <linux/mdev.h>
#include <linux/pci.h>

#undef  pr_fmt
#define pr_fmt(fmt)				"hymccp: " fmt

#define VERSION_STRING				"0.1"
#define DRIVER_AUTHOR				"HYGON Corporation"

#define MCCP_CLASS_NAME 			"hymccp"
#define MCCP_NAME				"hymccp"
#define MCCP_STRING_LEN				16

#define MCCP_CONFIG_SPACE_SIZE			0xff

#define MCCP_VFIO_PCI_OFFSET_SHIFT		40
#define MCCP_VFIO_PCI_OFFSET_TO_INDEX(off) \
				(off >> MCCP_VFIO_PCI_OFFSET_SHIFT)
#define MCCP_VFIO_PCI_INDEX_TO_OFFSET(index) \
				((u64)(index) << MCCP_VFIO_PCI_OFFSET_SHIFT)
#define MCCP_VFIO_PCI_OFFSET_MASK \
				(((u64)(1) << MCCP_VFIO_PCI_OFFSET_SHIFT) - 1)

#define MCCP_SHARE_IOC_TYPE			'C'
#define MCCP_SHARE_OP				0x01
#define MCCP_SHARE_OP_DMA_MAP			0x01
#define MCCP_SHARE_OP_DMA_UNMAP			0x02
#define MCCP_SHARE_OP_GET_ID			0x03

#define MCCP_SHARE_IOMMU_MAGIC			0x3d6a9c5728633b9e

/* fixed iova range for ccp dma. */
#define MCCP_DMA_IOVA_OFFSET			0
#define MCCP_DMA_IOVA_SIZE			(1ul << 30)

#define MCCP_INSTANCE_MAX			1024
#define MCCP_INSTANCE_OFFSET			8
#define MCCP_INSTANCE_MASK			(~((1u << MCCP_INSTANCE_OFFSET) - 1))
#define CCP_IOVA_MAX_SLOT			1024
#define MCCP_DEV_MAX				16
#define MCCP_DEV_QUEUE_MAX			8
#define MCCP_QUEUES_MAX				(MCCP_DEV_MAX * MCCP_DEV_QUEUE_MAX)
#define MCCP_QUEUE_NEED_INIT			0x01

struct mccp_shared_cfg {
	volatile unsigned int iova_slot[CCP_IOVA_MAX_SLOT];
	volatile unsigned int ccp_queue_state[MCCP_QUEUES_MAX];
	volatile unsigned int ccps_ref[MCCP_DEV_MAX];
	volatile unsigned int ccps_ref_lock;
	int rsvd2[15];
	volatile u64 qidx[MCCP_QUEUES_MAX];
	volatile unsigned int ccp_state[MCCP_DEV_MAX];
} __attribute__((__aligned__(PAGE_SIZE)));

struct mccp_dev_ctrl {
	unsigned char op;
	unsigned char rsvd[3];
	union {
		struct {
			unsigned long vaddr;
			unsigned long iova;
			unsigned long size;
		};
		unsigned int id;
	};
};

struct mccp_dma {
	struct list_head next;
	unsigned long vaddr;
	unsigned long iova;
	size_t size;
	struct page **pages;
	unsigned long npages;
};

struct mccp_iommu {
	unsigned long magic;
	struct mutex lock;
	struct pci_dev *pdev;
	unsigned long id;
	unsigned long ref;
};

static struct mccp_data {
	struct mccp_iommu iommu[MCCP_DEV_MAX];
	struct mutex lock;
	unsigned long bitmap;
	struct iommu_domain *domain;
	int prot;
	dma_addr_t dma_share_iova;
	size_t dma_share_size;
	unsigned long dma_share_ref;
	unsigned long mdev_ref;
	unsigned long ids[BITS_TO_LONGS(MCCP_INSTANCE_MAX)];
} mccp_data;

static struct mccp_share_cfg {
	long ref;
	struct mutex lock;
	void *vaddr;
	u64 size;
} mccp_share;

static struct mccp_dev {
	dev_t vd_devt;
	struct class *vd_class;
	struct cdev vd_cdev;
	struct device dev;
} mccp_dev;

struct mdev_region_info {
	u64 start;
	u64 phys_start;
	u32 size;
	u64 vfio_offset;
};

struct mdev_state {
	struct mutex ops_lock;
	struct mdev_device *mdev;
	struct mccp_iommu *iommu;
	struct mdev_region_info region_info[VFIO_PCI_NUM_REGIONS];
	struct list_head next;
	struct vfio_device_info dev_info;
};

static int mccp_iommu_alloc(struct pci_dev *pdev)
{
	unsigned long i;
	int ret = -EINVAL;
	uint32_t qmask = 0;
	void __iomem *resource2;

	resource2 = pci_iomap(pdev, 2, 0);
	qmask = ((uint32_t *)resource2)[0] & 0x1f;
	pci_iounmap(pdev, resource2);
	if (!qmask)
		return -ECANCELED;

	mutex_lock(&mccp_data.lock);

	i = find_first_zero_bit(&mccp_data.bitmap, MCCP_DEV_MAX);
	if (i != MCCP_DEV_MAX)
		bitmap_set(&mccp_data.bitmap, i, 1);

	if (iommu_capable(pdev->dev.bus, IOMMU_CAP_CACHE_COHERENCY))
		mccp_data.prot |= IOMMU_CACHE;

	mutex_unlock(&mccp_data.lock);

	if (i == MCCP_DEV_MAX)
		return -EINVAL;

	ret = iommu_attach_device(mccp_data.domain, &pdev->dev);
	if (ret) {
		mutex_lock(&mccp_data.lock);
		bitmap_clear(&mccp_data.bitmap, i, 1);
		mutex_unlock(&mccp_data.lock);
	} else {
		mutex_lock(&mccp_data.iommu[i].lock);
		mccp_data.iommu[i].pdev = pdev;
		mccp_data.iommu[i].id = i;
		mccp_data.iommu[i].ref = 0;
		mccp_data.iommu[i].magic = MCCP_SHARE_IOMMU_MAGIC;
		pci_set_drvdata(pdev, &mccp_data.iommu[i]);
		mutex_unlock(&mccp_data.iommu[i].lock);
	}

	return ret;
}

static void mccp_iommu_free(struct mccp_iommu *iommu)
{
	struct iommu_domain *domain;

	if (!iommu || iommu->magic != MCCP_SHARE_IOMMU_MAGIC)
		return;

	domain = iommu_get_domain_for_dev(&iommu->pdev->dev);

	mutex_lock(&iommu->lock);
	if (iommu->pdev && domain == mccp_data.domain)
		iommu_detach_device(domain, &iommu->pdev->dev);
	iommu->pdev = NULL;
	iommu->magic = 0;
	mutex_unlock(&iommu->lock);

	mutex_lock(&mccp_data.lock);
	if (iommu->id < MCCP_DEV_MAX)
		bitmap_clear(&mccp_data.bitmap, iommu->id, 1);
	mutex_unlock(&mccp_data.lock);
}

static int handle_pci_cfg_read(struct mdev_state *mdev_state, int offset,
				 __le32 *val, int count)
{
	u32 tmp_val = 0;
	int ret = -EINVAL;
	struct pci_dev *pdev = mdev_state->iommu->pdev;

	if (!mdev_state->mdev || !pdev) {
		pr_err("hymccp: invalid dev or pdev\n");
		return ret;
	}

	if (count == 1) {
		u8 tmp;
		ret = pci_user_read_config_byte(pdev, offset, &tmp);
		tmp_val = tmp;
	} else if (count ==2) {
		u16 tmp;
		ret = pci_user_read_config_word(pdev, offset, &tmp);
		tmp_val = tmp;
	} else if (count == 4)
		ret = pci_user_read_config_dword(pdev, offset, &tmp_val);

	*val = cpu_to_le32(tmp_val);

	return ret;
}

static int handle_pci_cfg_write(struct mdev_state *mdev_state, u16 offset,
				 u8 *buf, u32 count)
{
	u32 tmp_val = le32_to_cpu(*(u32 *)buf);
	int ret = -EINVAL;
	struct pci_dev *pdev = mdev_state->iommu->pdev;


	if (!mdev_state->mdev || !pdev) {
		pr_err("hymccp: invalid dev or pdev\n");
		return ret;
	}

	if (count == 1)
		ret = pci_user_write_config_byte(pdev, offset, tmp_val);
	else if (count == 2)
		ret = pci_user_write_config_word(pdev, offset, tmp_val);
	else if (count == 4)
		ret = pci_user_write_config_dword(pdev, offset, tmp_val);

	return ret;
}

static ssize_t mccp_access(struct mdev_device *mdev, u8 *buf, size_t count,
			   loff_t pos, bool is_write)
{
	struct mdev_state *mdev_state;
	unsigned int index;
	loff_t offset;
	int ret = 0;

	if (!mdev || !buf)
		return -EINVAL;

	mdev_state = mdev_get_drvdata(mdev);
	if (!mdev_state) {
		pr_err("%s mdev_state not found\n", __func__);
		return -EINVAL;
	}

	mutex_lock(&mdev_state->ops_lock);

	index = MCCP_VFIO_PCI_OFFSET_TO_INDEX(pos);
	offset = pos & MCCP_VFIO_PCI_OFFSET_MASK;
	switch (index) {
	case VFIO_PCI_CONFIG_REGION_INDEX:
		if (is_write)
			ret = handle_pci_cfg_write(mdev_state, offset, buf, count);
		else
			ret = handle_pci_cfg_read(mdev_state, offset, (__le32 *)buf, count);
		break;
	default:
		ret = -1;
	}

	if (!ret)
		ret = count;

	mutex_unlock(&mdev_state->ops_lock);

	return ret;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 13, 0)
static int mccp_create(struct mdev_device *mdev)
#else
static int mccp_create(struct kobject *kobj, struct mdev_device *mdev)
#endif
{
	struct mdev_state *mdev_state = NULL;
	unsigned long ref = -1ul;
	int n = -1;
	unsigned long *bitmap = &mccp_data.bitmap;
	struct mccp_iommu *iommu = mccp_data.iommu;
	int i;
	int ret = 0;

	if (!mdev)
		return -EINVAL;

	mdev_state = kzalloc(sizeof(struct mdev_state), GFP_KERNEL);
	if (mdev_state == NULL)
		return -ENOMEM;

	mutex_init(&mdev_state->ops_lock);

	mutex_lock(&mccp_data.lock);
	if (mccp_data.mdev_ref > 0) {
		mutex_unlock(&mccp_data.lock);
		ret = -EBUSY;
		goto exit;
	}

	for (i = 0; i < MCCP_DEV_MAX; i++) {
		if (test_bit(i, bitmap)) {
			if (ref > iommu[i].ref) {
				n = i;
				ref = iommu[i].ref;
			}
		}
	}

	if (n >= 0 && n < MCCP_DEV_MAX) {
		iommu[n].ref++;
		mdev_state->iommu = &iommu[n];
	} else
		ret = -EINVAL;
	mutex_unlock(&mccp_data.lock);

	if (ret)
		goto exit;

	mdev_state->mdev = mdev;
	mdev_set_drvdata(mdev, mdev_state);

	return 0;

exit:
	kfree(mdev_state);
	return ret;
}

static int mccp_remove(struct mdev_device *mdev)
{
	struct mdev_state *mdev_state = mdev_get_drvdata(mdev);
	int ret = 0;

	mutex_lock(&mccp_data.lock);
	if (mccp_data.mdev_ref > 0)
		ret = -EBUSY;
	else {
		mdev_state->iommu->ref--;
		mdev_set_drvdata(mdev, NULL);
		kfree(mdev_state);
	}
	mutex_unlock(&mccp_data.lock);

	return ret;
}

static int mccp_reset(struct mdev_device *mdev)
{
	struct mdev_state *mdev_state;

	if (!mdev)
		return -EINVAL;

	mdev_state = mdev_get_drvdata(mdev);
	if (!mdev_state)
		return -EINVAL;

	return 0;
}

static ssize_t mccp_read(struct mdev_device *mdev, char __user *buf,
			 size_t count, loff_t *ppos)
{
	unsigned int done = 0;
	int ret;

	while (count) {
		size_t filled;

		if (count >= 4 && !(*ppos % 4)) {
			u32 val;

			ret =  mccp_access(mdev, (u8 *)&val, sizeof(val), *ppos, false);
			if (ret <= 0)
				goto read_err;

			if (copy_to_user(buf, &val, sizeof(val)))
				goto read_err;

			filled = 4;
		} else if (count >= 2 && !(*ppos % 2)) {
			u16 val;

			ret = mccp_access(mdev, (u8 *)&val, sizeof(val), *ppos, false);
			if (ret <= 0)
				goto read_err;

			if (copy_to_user(buf, &val, sizeof(val)))
				goto read_err;

			filled = 2;
		} else {
			u8 val;

			ret = mccp_access(mdev, (u8 *)&val, sizeof(val), *ppos, false);
			if (ret <= 0)
				goto read_err;

			if (copy_to_user(buf, &val, sizeof(val)))
				goto read_err;

			filled = 1;
		}

		count -= filled;
		done += filled;
		*ppos += filled;
		buf += filled;
	}

	return done;

read_err:
	return -EFAULT;
}

static ssize_t mccp_write(struct mdev_device *mdev, const char __user *buf,
		   size_t count, loff_t *ppos)
{
	unsigned int done = 0;
	int ret;

	while (count) {
		size_t filled;

		if (count >= 4 && !(*ppos % 4)) {
			u32 val;

			if (copy_from_user(&val, buf, sizeof(val)))
				goto write_err;

			ret = mccp_access(mdev, (u8 *)&val, sizeof(val), *ppos, true);
			if (ret <= 0)
				goto write_err;

			filled = 4;
		} else if (count >= 2 && !(*ppos % 2)) {
			u16 val;

			if (copy_from_user(&val, buf, sizeof(val)))
				goto write_err;

			ret = mccp_access(mdev, (u8 *)&val, sizeof(val), *ppos, true);
			if (ret <= 0)
				goto write_err;

			filled = 2;
		} else {
			u8 val;

			if (copy_from_user(&val, buf, sizeof(val)))
				goto write_err;

			ret = mccp_access(mdev, (u8 *)&val, sizeof(val), *ppos, true);
			if (ret <= 0)
				goto write_err;

			filled = 1;
		}
		count -= filled;
		done += filled;
		*ppos += filled;
		buf += filled;
	}

	return done;
write_err:
	return -EFAULT;
}

static int mccp_get_region_info(struct mdev_device *mdev,
			 struct vfio_region_info *region_info,
			 u16 *cap_type_id, void **cap_type)
{
	unsigned int size = 0;
	u32 bar_index;
	struct mdev_state *mdev_state;
	struct pci_dev *pdev;

	if (!mdev)
		return -EINVAL;

	mdev_state = mdev_get_drvdata(mdev);
	if (!mdev_state)
		return -ENODEV;

	bar_index = region_info->index;
	if (bar_index >= VFIO_PCI_NUM_REGIONS)
		return -EINVAL;

	pdev = mdev_state->iommu->pdev;
	mutex_lock(&mdev_state->ops_lock);

	switch (bar_index) {
	case VFIO_PCI_CONFIG_REGION_INDEX:
		size = pdev->cfg_size;
		break;
	case VFIO_PCI_BAR0_REGION_INDEX ... VFIO_PCI_BAR5_REGION_INDEX:
		size = pci_resource_len(pdev, bar_index);
		break;
	default:
		size = 0;
		break;
	}

	mdev_state->region_info[bar_index].size = size;
	mdev_state->region_info[bar_index].vfio_offset =
		MCCP_VFIO_PCI_INDEX_TO_OFFSET(bar_index);

	region_info->size = size;
	region_info->offset = MCCP_VFIO_PCI_INDEX_TO_OFFSET(bar_index);
	region_info->flags = VFIO_REGION_INFO_FLAG_READ |
		VFIO_REGION_INFO_FLAG_WRITE;
	if (size >= PAGE_SIZE)
		region_info->flags |= VFIO_REGION_INFO_FLAG_MMAP;

	mutex_unlock(&mdev_state->ops_lock);
	return 0;
}

static int mccp_get_irq_info(struct mdev_device *mdev,
			     struct vfio_irq_info *irq_info)
{
	switch (irq_info->index) {
	case VFIO_PCI_INTX_IRQ_INDEX:
	case VFIO_PCI_MSI_IRQ_INDEX:
	case VFIO_PCI_REQ_IRQ_INDEX:
		break;

	default:
		return -EINVAL;
	}

	irq_info->flags = VFIO_IRQ_INFO_EVENTFD;
	irq_info->count = 1;

	if (irq_info->index == VFIO_PCI_INTX_IRQ_INDEX)
		irq_info->flags |= (VFIO_IRQ_INFO_MASKABLE |
				VFIO_IRQ_INFO_AUTOMASKED);
	else
		irq_info->flags |= VFIO_IRQ_INFO_NORESIZE;

	return 0;
}

static int mccp_get_device_info(struct mdev_device *mdev,
			 struct vfio_device_info *dev_info)
{
	dev_info->flags = VFIO_DEVICE_FLAGS_PCI;
	dev_info->num_regions = VFIO_PCI_NUM_REGIONS;
	dev_info->num_irqs = VFIO_PCI_NUM_IRQS;

	return 0;
}

static long mccp_ioctl(struct mdev_device *mdev, unsigned int cmd,
			unsigned long arg)
{
	int ret = 0;
	unsigned long minsz;
	struct mdev_state *mdev_state;

	if (!mdev)
		return -EINVAL;

	mdev_state = mdev_get_drvdata(mdev);
	if (!mdev_state)
		return -ENODEV;

	if (!mdev_state->iommu || !mdev_state->iommu->pdev)
		return -EIO;

	switch (cmd) {
	case VFIO_DEVICE_GET_INFO:
	{
		struct vfio_device_info info;

		minsz = offsetofend(struct vfio_device_info, num_irqs);

		if (copy_from_user(&info, (void __user *)arg, minsz))
			return -EFAULT;

		if (info.argsz < minsz)
			return -EINVAL;

		ret = mccp_get_device_info(mdev, &info);
		if (ret)
			return ret;

		memcpy(&mdev_state->dev_info, &info, sizeof(info));

		if (copy_to_user((void __user *)arg, &info, minsz))
			return -EFAULT;

		return 0;
	}
	case VFIO_DEVICE_GET_REGION_INFO:
	{
		struct vfio_region_info info;
		u16 cap_type_id = 0;
		void *cap_type = NULL;

		minsz = offsetofend(struct vfio_region_info, offset);

		if (copy_from_user(&info, (void __user *)arg, minsz))
			return -EFAULT;

		if (info.argsz < minsz)
			return -EINVAL;

		ret = mccp_get_region_info(mdev, &info, &cap_type_id,
					   &cap_type);
		if (ret)
			return ret;

		if (copy_to_user((void __user *)arg, &info, minsz))
			return -EFAULT;

		return 0;
	}

	case VFIO_DEVICE_GET_IRQ_INFO:
	{
		struct vfio_irq_info info;

		minsz = offsetofend(struct vfio_irq_info, count);

		if (copy_from_user(&info, (void __user *)arg, minsz))
			return -EFAULT;

		if ((info.argsz < minsz) ||
		    (info.index >= mdev_state->dev_info.num_irqs))
			return -EINVAL;

		ret = mccp_get_irq_info(mdev, &info);
		if (ret)
			return ret;

		if (copy_to_user((void __user *)arg, &info, minsz))
			return -EFAULT;

		return 0;
	}
	case VFIO_DEVICE_SET_IRQS:
	{
		return 0;
	}
	case VFIO_DEVICE_RESET:
		return mccp_reset(mdev);
	}
	return -ENOTTY;
}

static int mccp_open(struct mdev_device *mdev)
{
	struct mdev_state *mdev_state = mdev_get_drvdata(mdev);

	if (!mdev_state)
		return -ENODEV;

	if (!mdev_state->iommu || !mdev_state->iommu->pdev)
		return -EIO;

	mutex_lock(&mccp_data.lock);
	mccp_data.mdev_ref++;
	mutex_unlock(&mccp_data.lock);

	return 0;
}

static void mccp_close(struct mdev_device *mdev)
{
	mutex_lock(&mccp_data.lock);
	mccp_data.mdev_ref--;
	mutex_unlock(&mccp_data.lock);

	return;
}

static ssize_t mccp_dev_show(struct device *dev, struct device_attribute *attr,
			     char *buf)
{
	return sprintf(buf, "This is HYGON CCP device!\n");
}

static DEVICE_ATTR_RO(mccp_dev);

static struct attribute *ccp_dev_attrs[] = {
	&dev_attr_mccp_dev.attr,
	NULL,
};

static const struct attribute_group mccp_dev_group = {
	.name  = "hymccp_dev",
	.attrs = ccp_dev_attrs,
};

static const struct attribute_group *mccp_dev_groups[] = {
	&mccp_dev_group,
	NULL,
};

static ssize_t address_show(struct device *dev, struct device_attribute *attr,
			    char *buf)
{
	struct mdev_state *mdev_state;
	struct pci_dev *pdev;
	ssize_t size;

	if (mdev_from_dev(dev)) {
		mdev_state = mdev_get_drvdata(mdev_from_dev(dev));
		if (!mdev_state || !mdev_state->iommu)
			goto exit;

		mutex_lock(&mdev_state->iommu->lock);
		if (!mdev_state->iommu->pdev ||
		    mdev_state->iommu->magic != MCCP_SHARE_IOMMU_MAGIC) {
			mutex_unlock(&mdev_state->iommu->lock);
			goto exit;
		}
		pdev = mdev_state->iommu->pdev;
		size = sprintf(buf, "%04x:%02x:%02x.%x",
				pci_domain_nr(pdev->bus),
				pdev->bus->number,
				0x00ff & (pdev->devfn >> 8),
				0x00ff & pdev->devfn);
		mutex_unlock(&mdev_state->iommu->lock);
		return size;
	}

exit:
	return sprintf(buf, "\n");
}

static ssize_t id_show(struct device *dev, struct device_attribute *attr,
			    char *buf)
{
	struct mdev_state *mdev_state;
	ssize_t size;

	if (mdev_from_dev(dev)) {
		mdev_state = mdev_get_drvdata(mdev_from_dev(dev));
		if (!mdev_state || !mdev_state->iommu)
			goto exit;

		mutex_lock(&mdev_state->iommu->lock);
		if (!mdev_state->iommu->pdev ||
		    mdev_state->iommu->magic != MCCP_SHARE_IOMMU_MAGIC) {
			mutex_unlock(&mdev_state->iommu->lock);
			goto exit;
		}

		size = sprintf(buf, "%lu", mdev_state->iommu->id);
		mutex_unlock(&mdev_state->iommu->lock);
		return size;
	}

exit:
	return sprintf(buf, "\n");
}

static DEVICE_ATTR_RO(address);
static DEVICE_ATTR_RO(id);

static struct attribute *mdev_dev_attrs[] = {
	&dev_attr_address.attr,
	&dev_attr_id.attr,
	NULL,
};

static const struct attribute_group mdev_dev_group = {
	.name  = "vendor",
	.attrs = mdev_dev_attrs,
};

static const struct attribute_group *mccp_mdev_groups[] = {
	&mdev_dev_group,
	NULL,
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 16, 0)
static ssize_t
name_show(struct mdev_type *mtype, struct mdev_type_attribute *attr, char *buf)
#else
static ssize_t
name_show(struct kobject *kobj, struct device *dev, char *buf)
#endif
{
	return sprintf(buf, "%s\n", "hymccp");
}

static MDEV_TYPE_ATTR_RO(name);

static struct attribute *mccp_types_attrs[] = {
	&mdev_type_attr_name.attr,
	NULL,
};

static struct attribute_group mccp_type_group1 = {
	.name  = "1",
	.attrs = mccp_types_attrs,
};

static struct attribute_group *mccp_type_groups[] = {
	&mccp_type_group1,
	NULL,
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 19, 0)
static void mccp_mmap_open(struct vm_area_struct *vma)
{
	zap_vma_ptes(vma, vma->vm_start, vma->vm_end - vma->vm_start);
}

static void mccp_mmap_close(struct vm_area_struct *vma)
{
}

static vm_fault_t mccp_mmap_fault(struct vm_fault *vmf)
{
	vm_fault_t ret = VM_FAULT_NOPAGE;
	struct vm_area_struct *vma = vmf->vma;
	if (remap_pfn_range(vma, vma->vm_start, vma->vm_pgoff,
			vma->vm_end - vma->vm_start, pgprot_decrypted(vma->vm_page_prot)))
		ret= VM_FAULT_SIGBUS;
	return ret;
}

static const struct vm_operations_struct mccp_mmap_ops = {
	.open = mccp_mmap_open,
	.close = mccp_mmap_close,
	.fault = mccp_mmap_fault,
};

static int mccp_mmap(struct mdev_device *mdev, struct vm_area_struct *vma)
{
	unsigned int index;
	struct mdev_state *mds = mdev_get_drvdata(mdev);
	struct pci_dev *pdev = mds->iommu->pdev;

	index = vma->vm_pgoff >> (40 - PAGE_SHIFT);
	vma->vm_private_data = mdev;
	vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
	vma->vm_pgoff = pci_resource_start(pdev, index) >> PAGE_SHIFT;
	vma->vm_flags |= VM_IO | VM_PFNMAP | VM_DONTEXPAND | VM_DONTDUMP;
	vma->vm_ops = &mccp_mmap_ops;

	return 0;
}
#else
static int mccp_mmap(struct mdev_device *mdev, struct vm_area_struct *vma)
{
	unsigned int index;
	struct mdev_state *mds = mdev_get_drvdata(mdev);
	struct pci_dev *pdev = mds->iommu->pdev;

	index = vma->vm_pgoff >> (40 - PAGE_SHIFT);
	vma->vm_private_data = mdev;
	vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
	vma->vm_pgoff = pci_resource_start(pdev, index) >> PAGE_SHIFT;
	vma->vm_flags |= VM_IO | VM_PFNMAP | VM_DONTEXPAND | VM_DONTDUMP;

	return remap_pfn_range(vma, vma->vm_start, vma->vm_pgoff,
			       vma->vm_end - vma->vm_start,
			       pgprot_decrypted(vma->vm_page_prot));
}
#endif

static const struct mdev_parent_ops mccp_mdev_fops = {
	.owner			= THIS_MODULE,
	.dev_attr_groups	= mccp_dev_groups,
	.mdev_attr_groups	= mccp_mdev_groups,
	.supported_type_groups	= mccp_type_groups,
	.create 		= mccp_create,
	.remove			= mccp_remove,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 15, 0)
	.open_device		= mccp_open,
	.close_device		= mccp_close,
#else
	.open			= mccp_open,
	.release		= mccp_close,
#endif
	.read			= mccp_read,
	.write			= mccp_write,
	.ioctl			= mccp_ioctl,
	.mmap			= mccp_mmap,
};

struct mccp_private {
	struct list_head head;
	struct mutex lock;
	unsigned long pages_locked;
	unsigned int id;
};

static int mccp_share_open(struct inode *inode, struct file *file)
{
	int ret = 0;
	struct mccp_private *private;
	unsigned int id;

	private = kzalloc(sizeof(*private), GFP_KERNEL);
	if (!private)
		return -ENOMEM;

	mutex_lock(&mccp_data.lock);
	bitmap_set(mccp_data.ids, 0, 1);
	id = (unsigned int)find_first_zero_bit(mccp_data.ids, MCCP_INSTANCE_MAX);
	if (id < MCCP_INSTANCE_MAX)
		bitmap_set(mccp_data.ids, id, 1);
	mutex_unlock(&mccp_data.lock);

	if (id >= MCCP_INSTANCE_MAX) {
		kfree(private);
		return -EBUSY;
	}

	private->id = id << MCCP_INSTANCE_OFFSET;
	mutex_lock(&mccp_share.lock);
	mccp_share.ref++;
	mutex_unlock(&mccp_share.lock);

	file->private_data = private;
	INIT_LIST_HEAD(&private->head);
	mutex_init(&private->lock);

	return ret;
}

struct page **mccp_pin_memory(struct mccp_private *private, unsigned long uaddr,
			unsigned long ulen, unsigned long *n)
{
	unsigned long npages, size;
	int npinned;
	unsigned long locked, lock_limit;
	struct page **pages;
	unsigned long first, last;

	if (ulen == 0 || uaddr + ulen < uaddr)
		return NULL;

	first = (uaddr & PAGE_MASK) >> PAGE_SHIFT;
	last = ((uaddr + ulen - 1) & PAGE_MASK) >> PAGE_SHIFT;
	npages = (last - first + 1);

	locked = private->pages_locked + npages;
	lock_limit = rlimit(RLIMIT_MEMLOCK) >> PAGE_SHIFT;
	if (locked > lock_limit && !capable(CAP_IPC_LOCK)) {
		pr_err("%lu locked pages exceed the lock limit of %lu.\n", locked, lock_limit);
		return NULL;
	}

	if (WARN_ON_ONCE(npages > INT_MAX))
		return NULL;

	size = npages * sizeof(struct page *);
	if (size > PAGE_SIZE)
		pages = vmalloc(size);
	else
		pages = kmalloc(size, GFP_KERNEL);

	if (!pages)
		return NULL;

	/* Pin the user virtual address. */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
	npinned = pin_user_pages_fast(uaddr, npages, FOLL_WRITE, pages);
#else
	npinned = get_user_pages_fast(uaddr, npages, FOLL_WRITE, pages);
#endif
	if (npinned != npages) {
		pr_err("Failure locking %lu pages.\n", npages);
		goto err;
	}

	*n = npages;
	private->pages_locked = locked;
	return pages;

err:
	if (npinned > 0)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
		unpin_user_pages(pages, npinned);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0)
		release_pages(pages, npinned);
#else
		release_pages(pages, npinned, true);
#endif
	kvfree(pages);
	return NULL;
}

static void mccp_unpin_memory(struct mccp_private *private, struct page **pages,
			     unsigned long npages)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
	unpin_user_pages(pages, npages);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0)
	release_pages(pages, npages);
#else
	release_pages(pages, npages, true);
#endif
	kvfree(pages);
	private->pages_locked -= npages;
}

static inline int is_dma_share(dma_addr_t dma_iova, size_t dma_size)
{
	if (dma_iova >= MCCP_DMA_IOVA_OFFSET &&
	    dma_iova + dma_size <= MCCP_DMA_IOVA_OFFSET + MCCP_DMA_IOVA_SIZE)
		return 1;
	else
		return 0;
}

static int mccp_add_dma_share_unsafe(dma_addr_t dma_iova, size_t dma_size)
{
	int ret = 0;

	if (!is_dma_share(dma_iova, dma_size))
		return 0;

	if (!mccp_data.dma_share_size) {
		mccp_data.dma_share_iova = dma_iova;
		mccp_data.dma_share_size = dma_size;
	}

	if (dma_iova != mccp_data.dma_share_iova ||
	    dma_size != mccp_data.dma_share_size)
		ret = -EINVAL;
	else
		mccp_data.dma_share_ref++;

	return ret;
}

static int mccp_unmap_dma_share_unsafe(dma_addr_t dma_iova, size_t dma_size)
{
	if (!is_dma_share(dma_iova, dma_size))
		return -EINVAL;

	if (mccp_data.dma_share_size) {
		if (mccp_data.dma_share_iova == dma_iova &&
		    mccp_data.dma_share_size == dma_size)
			mccp_data.dma_share_ref--;

		if (mccp_data.dma_share_ref == 0) {
			iommu_unmap(mccp_data.domain, mccp_data.dma_share_iova,
				mccp_data.dma_share_size);
			mccp_data.dma_share_size = 0;
		}
	}

	return 0;
}

static int mccp_iommu_iova_check_unsafe(dma_addr_t dma_iova, size_t dma_size,
				phys_addr_t phys_addr,
				struct iommu_domain *domain)
{
	dma_addr_t iova;
	int ret = 0;
	size_t mapped = 0;

	iova = dma_iova;
	while (iova < dma_iova + dma_size) {
		phys_addr_t phys;

		phys = iommu_iova_to_phys(domain, iova);
		if (phys) {
			if (phys_addr != phys) {
				ret = -1;
				break;
			}
			mapped += PAGE_SIZE;
		}
		iova += PAGE_SIZE;
		phys_addr += PAGE_SIZE;
	}

	if (ret == 0 && mapped == dma_size)
		ret = 1;

	return ret;
}

static unsigned long get_num_contig_pages(unsigned long idx,
				struct page **inpages, unsigned long npages)
{
	unsigned long paddr, next_paddr;
	unsigned long i = idx + 1, pages = 1;

	/* find the number of contiguous pages starting from idx */
	paddr = page_to_phys(inpages[idx]);
	while (i < npages) {
		next_paddr = page_to_phys(inpages[i++]);
		if ((paddr + PAGE_SIZE) == next_paddr) {
			pages++;
			paddr = next_paddr;
			continue;
		}
		break;
	}

	return pages;
}

static struct mccp_dma *mccp_find_dma(struct mccp_private *private,
				      dma_addr_t start, size_t size)
{
	struct mccp_dma *dma, *tmp;

	list_for_each_entry_safe(dma, tmp, &private->head, next) {
		if (dma->iova <= start &&
		    dma->iova + dma->size >= start + size)
			return dma;
	}

	return NULL;
}

/*
 * Turns out AMD IOMMU has a page table bug where it won't map large pages
 * to a region that previously mapped smaller pages.  This should be fixed
 * soon, so this is just a temporary workaround to break mappings down into
 * PAGE_SIZE.  Better to map smaller pages than nothing.
 */
static int map_try_harder(struct iommu_domain *domain, dma_addr_t iova,
			  unsigned long pfn, long npage, int prot)
{
	long i;
	int ret = 0;

	for (i = 0; i < npage; i++, pfn++, iova += PAGE_SIZE) {
		ret = iommu_map(domain, iova,
				(phys_addr_t)pfn << PAGE_SHIFT,
				PAGE_SIZE, prot);
		if (ret)
			break;
	}

	for (; i < npage && i > 0; i--, iova -= PAGE_SIZE)
		iommu_unmap(domain, iova, PAGE_SIZE);

	return ret;
}

static int mccp_iommu_map(struct mccp_private *private, unsigned long vaddr,
			dma_addr_t dma_iova, size_t dma_size)
{
	struct mccp_dma *dma;
	struct page **pages;
	unsigned long n, i, npages;
	dma_addr_t iova, iova_end, iova_next;
	int ret = 0;
	size_t mapped_size = 0;
	size_t iova_size = dma_size;

	if (!dma_size || (vaddr | dma_iova | dma_size) & (PAGE_SIZE - 1))
		return -EINVAL;

	if (mccp_find_dma(private, dma_iova, dma_size))
		return 0;

	dma = kzalloc(sizeof(*dma), GFP_KERNEL);
	if (!dma)
		return -ENOMEM;

	pages = mccp_pin_memory(private, vaddr, dma_size, &n);
	if (!pages) {
		ret = -ENOMEM;
		goto pin_fail;
	}

	dma->vaddr = vaddr;
	dma->iova = dma_iova;
	dma->pages = pages;
	dma->size = dma_size;
	dma->npages = n;

	iova = dma_iova;
	iova_end = dma_iova + dma_size;
	iova_size = dma_size;

	mutex_lock(&mccp_data.lock);
	for (i = 0; iova < iova_end && i < n; iova = iova_next, i += npages) {
		int len;
		phys_addr_t phys;

		npages = get_num_contig_pages(i, pages, n);

		len = min_t(size_t, (npages * PAGE_SIZE), iova_size);
		phys = page_to_phys(pages[i]);

		iova_size -= len;
		iova_next = iova + len;

		ret = mccp_iommu_iova_check_unsafe(iova, len, phys, mccp_data.domain);
		if (ret < 0) {
			ret = -EBUSY;
			goto map_fail;
		} else if (ret > 0) {
			ret = 0;
			continue;
		}

		ret = iommu_map(mccp_data.domain, iova, phys, len, mccp_data.prot);
		if (ret) {
			if (ret == -EBUSY)
				ret = map_try_harder(mccp_data.domain, iova,
						     phys >> PAGE_SHIFT,
						     len >> PAGE_SHIFT,
						     mccp_data.prot);
			if (ret)
				goto map_fail;
		}
		mapped_size += len;
		cond_resched();
	}

	ret = mccp_add_dma_share_unsafe(dma_iova, dma_size);
	if (ret)
		goto map_fail;

	mutex_unlock(&mccp_data.lock);
	list_add(&dma->next, &private->head);
	return 0;
map_fail:
	if (mapped_size)
		iommu_unmap(mccp_data.domain, dma_iova, mapped_size);
	mutex_unlock(&mccp_data.lock);
	mccp_unpin_memory(private, pages, n);
pin_fail:
	kfree(dma);
	return ret;
}


void mccp_iommu_unmap(struct mccp_private *private)
{
	struct mccp_dma *dma, *tmp;
	struct iommu_domain *domain;

	domain = mccp_data.domain;

	mutex_lock(&mccp_data.lock);
	list_for_each_entry_safe(dma, tmp, &private->head, next) {
		if (mccp_unmap_dma_share_unsafe(dma->iova, dma->size))
			iommu_unmap(domain, dma->iova, dma->size);
		mccp_unpin_memory(private, dma->pages, dma->npages);
		cond_resched();
		list_del(&dma->next);
		kfree(dma);
	}
	mutex_unlock(&mccp_data.lock);
}

static long mccp_share_ioctl(struct file *file, unsigned int ioctl, unsigned long arg)
{
	struct mccp_dev_ctrl dev_ctrl;
	unsigned int cmd_id;
	unsigned int len;
	int ret= 0;
	struct mccp_private *private = file->private_data;

	if (_IOC_TYPE(ioctl) != MCCP_SHARE_IOC_TYPE)
		return -EINVAL;

	cmd_id = _IOC_NR(ioctl);
	len = _IOC_SIZE(ioctl);

	if (cmd_id != MCCP_SHARE_OP)
		return -EINVAL;

	if (len != sizeof(dev_ctrl))
		return -EINVAL;

	if (copy_from_user(&dev_ctrl, (void __user*)arg, sizeof(dev_ctrl)))
		return -EINVAL;

	mutex_lock(&private->lock);
	switch(dev_ctrl.op) {
	case MCCP_SHARE_OP_DMA_MAP:
		ret = mccp_iommu_map(private, dev_ctrl.vaddr, dev_ctrl.iova, dev_ctrl.size);
		break;
	case MCCP_SHARE_OP_DMA_UNMAP:
		mccp_iommu_unmap(private);
		ret = 0;
		break;
	case MCCP_SHARE_OP_GET_ID:
		dev_ctrl.id = private->id;
		if (copy_to_user((void __user*)arg, &dev_ctrl, sizeof(dev_ctrl)))
			ret = -EINVAL;
		else
			ret = 0;
		break;
	default:
		ret = 0;
		break;
	}
	mutex_unlock(&private->lock);

	return ret;
}

static int mccp_share_close(struct inode *inode, struct file *file)
{
	struct mccp_private *private = file->private_data;
	struct mccp_shared_cfg *cfg = mccp_share.vaddr;
	int i = 0;
	unsigned int id;

	id = private->id >> MCCP_INSTANCE_OFFSET;

	mutex_lock(&mccp_share.lock);

	if (private->id == (MCCP_INSTANCE_MASK & cfg->ccps_ref_lock))
		cfg->ccps_ref_lock = 0;

	for (i = 0; i < MCCP_DEV_MAX; i++)
		if (private->id == (MCCP_INSTANCE_MASK & cfg->ccp_state[i]))
			cfg->ccp_state[i] = 0;

	for (i = 0; i < MCCP_QUEUES_MAX; i++)
		if (private->id == (MCCP_INSTANCE_MASK & cfg->ccp_queue_state[i]))
			cfg->ccp_queue_state[i] = MCCP_QUEUE_NEED_INIT;

	for (i = 0; i < CCP_IOVA_MAX_SLOT; i++)
		if (private->id == (MCCP_INSTANCE_MASK & cfg->iova_slot[i]))
			cfg->iova_slot[i] = 0;

	mccp_share.ref--;
	if (!mccp_share.ref && mccp_share.vaddr)
		memset(mccp_share.vaddr, 0x00, mccp_share.size);

	mutex_unlock(&mccp_share.lock);

	mutex_lock(&mccp_data.lock);
	if (id < MCCP_INSTANCE_MAX)
		bitmap_clear(mccp_data.ids, id, 1);
	mutex_unlock(&mccp_data.lock);

	mutex_lock(&private->lock);
	mccp_iommu_unmap(private);
	mutex_unlock(&private->lock);

	kfree(private);

	return 0;
}

static int mccp_share_mmap(struct file *file, struct vm_area_struct *vma)
{
	int ret;

	mutex_lock(&mccp_share.lock);
	if (unlikely(!mccp_share.vaddr)) {
		mccp_share.size = (vma->vm_end - vma->vm_start);
		mccp_share.vaddr = kzalloc(mccp_share.size, GFP_KERNEL);
	}

	if (!mccp_share.vaddr) {
		ret = -ENOMEM;
		goto exit;
	}

	if (mccp_share.size != (vma->vm_end - vma->vm_start)) {
		ret = -EINVAL;
		pr_err("invaid mccp share size\n");
		goto exit;
	}

	ret = remap_pfn_range(vma, vma->vm_start,
				virt_to_phys(mccp_share.vaddr) >> PAGE_SHIFT,
				mccp_share.size,
				vma->vm_page_prot);
exit:
	mutex_unlock(&mccp_share.lock);
	return ret;
}

static const struct file_operations mccp_share_fops = {
	.owner		= THIS_MODULE,
	.open		= mccp_share_open,
	.release	= mccp_share_close,
	.mmap		= mccp_share_mmap,
	.unlocked_ioctl	= mccp_share_ioctl,
};

static struct miscdevice mccp_misc = {
	.minor = MISC_DYNAMIC_MINOR,
	.name = "mccp_share",
	.fops = &mccp_share_fops,
};

static int mccp_share_init(void)
{
	int i;
	int ret;

	memset(&mccp_data, 0x00, sizeof(mccp_data));
	mutex_init(&mccp_data.lock);

	for (i = 0; i < MCCP_DEV_MAX; i++)
		mutex_init(&mccp_data.iommu[i].lock);

	ret = misc_register(&mccp_misc);
	if (!ret) {
		mccp_data.domain = iommu_domain_alloc(&pci_bus_type);
		if (!mccp_data.domain) {
			misc_deregister(&mccp_misc);
			if (!pci_bus_type.iommu_ops) {
				pr_err("iommu is disabled\n");
				return -ENODEV;
			}
			return -ENOMEM;
		}
		mccp_data.prot = IOMMU_READ | IOMMU_WRITE;
	}

	return ret;
}

static void mccp_share_exit(void)
{
	int i;
	struct mccp_iommu *iommu;
	struct iommu_domain *domain;
	struct pci_dev *pdev;

	mutex_lock(&mccp_data.lock);
	for (i = 0; i < MCCP_DEV_MAX; i++) {
		iommu = &mccp_data.iommu[i];
		pdev = iommu->pdev;
		if (pdev) {
			domain = iommu_get_domain_for_dev(&pdev->dev);
			if (domain == mccp_data.domain)
				iommu_detach_device(domain, &pdev->dev);
		}
	}
	mutex_unlock(&mccp_data.lock);

	if (mccp_data.domain)
		iommu_domain_free(mccp_data.domain);

	if (mccp_share.vaddr)
		kfree(mccp_share.vaddr);

	misc_deregister(&mccp_misc);
}

static int mccp_pci_probe(struct pci_dev *pdev, const struct pci_device_id *id)
{
	return mccp_iommu_alloc(pdev);
}

static void mccp_pci_remove(struct pci_dev *pdev)
{
	struct mccp_iommu *iommu;

	iommu = pci_get_drvdata(pdev);
	if (iommu)
		mccp_iommu_free(iommu);
	pci_set_drvdata(pdev, NULL);

	return;
}

static struct pci_device_id mccp_fb_pci_table[] = {
	{
		.vendor	   = 0x1d94,
		.device	   = 0x1468,
		.subvendor = 0x1d94,
		.subdevice = 0x1468,
	}, {
		.vendor	   = 0x1d94,
		.device	   = 0x1456,
		.subvendor = 0x1d94,
		.subdevice = 0x1456,
	}, {
		/* end of list */
	}
};

static struct pci_driver mccp_pci_driver = {
	.name = "hymccp",
	.id_table = mccp_fb_pci_table,
	.probe = mccp_pci_probe,
	.remove = mccp_pci_remove,
};

static const struct file_operations mccp_vd_fops = {
	.owner = THIS_MODULE,
};

static void mccp_device_release(struct device *dev)
{
	dev_dbg(dev, "hymccp: released\n");
}

#define CPUID_VENDOR_HygonGenuine_ebx	0x6f677948
#define CPUID_VENDOR_HygonGenuine_ecx	0x656e6975
#define CPUID_VENDOR_HygonGenuine_edx	0x6e65476e

static inline void _cpuid(unsigned int *eax, unsigned int *ebx,
			unsigned int *ecx, unsigned int *edx)
{
	asm volatile("cpuid"
	    : "=a" (*eax),
	      "=b" (*ebx),
	      "=c" (*ecx),
	      "=d" (*edx)
	    : "0" (*eax), "2" (*ecx)
	    : "memory");
}

static int __init mccp_dev_init(void)
{
	int ret = 0;
	u32 vendor_ebx = 0;
	u32 vendor_ecx = 0;
	u32 vendor_edx = 0;
	u32 vendor_eax = 0;

	_cpuid(&vendor_eax, &vendor_ebx, &vendor_ecx, &vendor_edx);

	/* HygonGenuine */
	if (!(vendor_ebx == CPUID_VENDOR_HygonGenuine_ebx &&
	      vendor_ecx == CPUID_VENDOR_HygonGenuine_ecx &&
	      vendor_edx == CPUID_VENDOR_HygonGenuine_edx)) {
		pr_err("Not hygon hardware\n");
		return -1;
	}

	memset(&mccp_dev, 0, sizeof(mccp_dev));

	ret = alloc_chrdev_region(&mccp_dev.vd_devt, 0, MINORMASK + 1,
				  MCCP_NAME);

	if (ret < 0) {
		pr_err("Error: failed to register mccp_dev, err:%d\n", ret);
		return ret;
	}

	cdev_init(&mccp_dev.vd_cdev, &mccp_vd_fops);
	cdev_add(&mccp_dev.vd_cdev, mccp_dev.vd_devt, MINORMASK + 1);

	mccp_dev.vd_class = class_create(THIS_MODULE, MCCP_CLASS_NAME);

	if (IS_ERR(mccp_dev.vd_class)) {
		pr_err("Error: failed to register mccp_dev class\n");
		ret = PTR_ERR(mccp_dev.vd_class);
		goto failed1;
	}

	mccp_dev.dev.class = mccp_dev.vd_class;
	mccp_dev.dev.release = mccp_device_release;
	dev_set_name(&mccp_dev.dev, "%s", MCCP_NAME);

	ret = device_register(&mccp_dev.dev);
	if (ret)
		goto failed2;

	ret = mdev_register_device(&mccp_dev.dev, &mccp_mdev_fops);
	if (ret)
		goto failed3;

	ret = mccp_share_init();
	if (ret)
		goto failed4;

	memset(&mccp_share, 0x00, sizeof(mccp_share));
	mutex_init(&mccp_share.lock);

	ret = pci_register_driver(&mccp_pci_driver);
	if (ret)
		goto failed5;

	goto all_done;

failed5:
	mccp_share_exit();

failed4:
	mdev_unregister_device(&mccp_dev.dev);

failed3:
	device_unregister(&mccp_dev.dev);

failed2:
	class_destroy(mccp_dev.vd_class);

failed1:
	cdev_del(&mccp_dev.vd_cdev);
	unregister_chrdev_region(mccp_dev.vd_devt, MINORMASK + 1);

all_done:
	return ret;
}

static void __exit mccp_dev_exit(void)
{
	mccp_share_exit();

	mccp_dev.dev.bus = NULL;
	mdev_unregister_device(&mccp_dev.dev);

	device_unregister(&mccp_dev.dev);
	cdev_del(&mccp_dev.vd_cdev);
	unregister_chrdev_region(mccp_dev.vd_devt, MINORMASK + 1);
	class_destroy(mccp_dev.vd_class);
	mccp_dev.vd_class = NULL;

	pci_unregister_driver(&mccp_pci_driver);
}

module_init(mccp_dev_init)
module_exit(mccp_dev_exit)

MODULE_LICENSE("GPL v2");
MODULE_VERSION(VERSION_STRING);
MODULE_AUTHOR(DRIVER_AUTHOR);

