#include <drm/drm_pci.h>
#include "inspur_drm_drv.h"
#include "inspur_drm_regs.h"

void colorcur2monocur(void *data, void *out)
{
    unsigned int *col = (unsigned int *)data;
    unsigned char *mono = (unsigned char *)out;
    unsigned char pixel = 0;
    char bit_values;

    int i;
    for (i = 0; i < 64 * 64; i++)
    {
        if (*col >> 24 < 0xe0)
        {
            bit_values = 0;
        }
        else
        {
            int val = *col & 0xff;

            if (val < 0x80)
            {
                bit_values = 1;
            }
            else
            {
                bit_values = 2;
            }
        }
        col++;
        /* Copy bits into cursor byte */
        switch (i & 3)
        {
        case 0:
            pixel = bit_values;
            break;

        case 1:
            pixel |= bit_values << 2;
            break;

        case 2:
            pixel |= bit_values << 4;
            break;

        case 3:
            pixel |= bit_values << 6;
            *mono = pixel;
            mono++;
            pixel = 0;
            break;
        }
    }
}


static int inspur_bo_reserve(struct inspur_bo *bo, bool no_wait)
{
	int ret;

	ret = ttm_bo_reserve(&bo->bo, true, no_wait, NULL);
	if (ret) {
		if (ret != -ERESTARTSYS && ret != -EBUSY)
			DRM_ERROR("reserve failed %p\n", bo);
        return ret;
    }
    return 0;
}

static void inspur_bo_unreserve(struct inspur_bo *bo)
{
	ttm_bo_unreserve(&bo->bo);
}

static void inspur_hide_cursor(struct inspur_drm_private *priv)
{
    unsigned int reg = readl(priv->mmio + INSPUR_HWC_ADDRESS);
    reg &= ~INSPUR_HWC_ADDRESS_ENABLE_MASK;
    writel(reg, priv->mmio + INSPUR_HWC_ADDRESS);
    if (priv->cursor.cursor_1->pin_count)
        inspur_bo_unpin(priv->cursor.cursor_1);
    if (priv->cursor.cursor_2->pin_count)
        inspur_bo_unpin(priv->cursor.cursor_2);
}

int inspur_crtc_cursor_set(
            struct drm_crtc *crtc,
            struct drm_file *file_priv,
            uint32_t handle,
            uint32_t width,
            uint32_t height
)
{
    struct drm_device *dev = crtc->dev;
    struct inspur_drm_private *priv = (struct inspur_drm_private *)dev->dev_private;
    struct inspur_bo *cursor_next = 0;
    struct drm_gem_object *obj;
    struct inspur_bo *bo = NULL;
    uint64_t gpu_addr;
    int ret;

    if (!priv->cursor.cursor_1 || !priv->cursor.cursor_2) {
        unsigned int reg = readl(priv->mmio + INSPUR_HWC_ADDRESS);
        reg &= ~INSPUR_HWC_ADDRESS_ENABLE_MASK;
        writel(reg, priv->mmio + INSPUR_HWC_ADDRESS);
        return -ENOTSUPP;
    }

    if ((width > 64 || height > 64) && handle) {
        unsigned int reg = readl(priv->mmio + INSPUR_HWC_ADDRESS);
        reg &= ~INSPUR_HWC_ADDRESS_ENABLE_MASK;
        writel(reg, priv->mmio + INSPUR_HWC_ADDRESS);
        return -EINVAL;
    }

    if (!handle || !file_priv) {
        inspur_hide_cursor(priv);
        return 0;
    }

    if (priv->cursor.cursor_current == priv->cursor.cursor_1)
        cursor_next = priv->cursor.cursor_2;
    else
        cursor_next = priv->cursor.cursor_1;

    obj = drm_gem_object_lookup(file_priv, handle);
    if (!obj)
        return -ENOENT;
    bo = gem_to_inspur_bo(obj);
    ret = inspur_bo_reserve(bo, true);
    if (ret) {
        dev_err(&dev->pdev->dev, "failed to reserve user bo\n");
        goto err_drm_gem_object_put_unlocked;
    }
    ret = inspur_bo_pin(bo, 0, NULL);
    if (ret) {
        dev_err(&dev->pdev->dev, "failed to lock user bo");
        goto err_drm_unreserve_src;
    }
    ret = ttm_bo_kmap(&bo->bo, 0, bo->bo.num_pages, &bo->kmap);
    if (ret) {
        dev_err(&dev->pdev->dev, "failed to kmap user buffer updates\n");
        goto err_drm_unpin_src;
    }

    ret = inspur_bo_reserve(cursor_next, true);
    if (ret) {
        goto err_drm_unmap_src;
    }
    ret = inspur_bo_pin(cursor_next, TTM_PL_FLAG_VRAM, &gpu_addr);
    if (ret) {
        goto err_drm_unreserve_dst;
    }


    //TODO: copy the cursor into hwcursor buffer
    DRM_DEBUG_KMS("set cursor %d - %dx%d\n", handle, width, height);
    colorcur2monocur(bo->kmap.virtual, priv->fb_map + gpu_addr);

    writel(0xffff0000, priv->mmio + INSPUR_HWC_COLOR_12);
    writel(0xffff, priv->mmio + INSPUR_HWC_COLOR_3);
    writel(INSPUR_FIELD(INSPUR_HWC_ADDRESS_ENABLE, 1) |
           INSPUR_FIELD(INSPUR_HWC_ADDRESS_ADDRESS, gpu_addr),
           priv->mmio + INSPUR_HWC_ADDRESS);

    if (priv->cursor.cursor_current)
        inspur_bo_unpin(priv->cursor.cursor_current);
    priv->cursor.cursor_current = cursor_next;

    inspur_bo_unreserve(cursor_next);
    ttm_bo_kunmap(&bo->kmap);
    inspur_bo_unpin(bo);
    inspur_bo_unreserve(bo);
    drm_gem_object_put_unlocked(obj);

    return 0;

err_drm_unreserve_dst:
    inspur_bo_unreserve(cursor_next);
err_drm_unmap_src:
    ttm_bo_kunmap(&bo->kmap);
err_drm_unpin_src:
    inspur_bo_unpin(bo);
err_drm_unreserve_src:
    inspur_bo_unreserve(bo);
err_drm_gem_object_put_unlocked:
    drm_gem_object_put_unlocked(obj);

    return ret;
}

int inspur_crtc_cursor_move(struct drm_crtc *crtc, int x, int y)
{
    struct inspur_drm_private *priv __attribute__((unused)) = (struct inspur_drm_private *)crtc->dev->dev_private;

    DRM_DEBUG_KMS("(%d, %d)\n", x, y);
    writel(INSPUR_FIELD(INSPUR_HWC_LOCATION_X, x) | INSPUR_FIELD(INSPUR_HWC_LOCATION_Y, y),
        priv->mmio + INSPUR_HWC_LOCATION);

    return 0;
}

#define HW_FLAG_OFFSET 0x01ffff00
#define HW_FLAG_ENABLE 0x1bd40750
unsigned char getKVMHWCursorSetting(struct inspur_drm_private *priv)
{
    unsigned int value = *(unsigned int*)(priv->fb_map + HW_FLAG_OFFSET);
    DRM_DEBUG_KMS("HW_FLAG = %x\n", value);
    if (value == HW_FLAG_ENABLE)
        return 1;
    else
        return 0;
}
