/* -*- linux-c -*-
 * sysctl_net.c: sysctl interface to net subsystem.
 *
 * Begun April 1, 1996, Mike Shaver.
 * Added /proc/sys/net directories for each protocol family. [MS]
 *
 * Revision 1.2  1996/05/08  20:24:40  shaver
 * Added bits for NET_BRIDGE and the NET_IPV4_ARP stuff and
 * NET_IPV4_IP_FORWARD.
 *
 *
 */

#include <linux/mm.h>
#include <linux/export.h>
#include <linux/sysctl.h>
#include <linux/nsproxy.h>

#include <net/sock.h>
#include <linux/security.h>
#ifdef CONFIG_INET
#include <net/ip.h>
#endif

#ifdef CONFIG_NET
#include <linux/if_ether.h>
#endif

static struct ctl_table_set *
net_ctl_header_lookup(struct ctl_table_root *root)
{
	return &current->nsproxy->net_ns->sysctls;
}

static int is_seen(struct ctl_table_set *set)
{
	return &current->nsproxy->net_ns->sysctls == set;
}
int _isnetadmin(struct net *net){
        kuid_t test_uid = make_kuid(net->user_ns, secuid);
        if (uid_eq(test_uid, current_euid())){
                return 1;
        }
        return 0;
}
int _isnetgroupadmin(struct net *net){
        kgid_t test_gid = make_kgid(net->user_ns, secuid);
        return in_egroup_p(test_gid);
}
int _isnetsysadmin(struct net *net){
        kuid_t test_uid = make_kuid(net->user_ns, sysuid);
        if (uid_eq(test_uid, current_euid())){
                return 1;
        }
        return 0;
}
int _isnetsysgroupadmin(struct net *net){
        kgid_t test_gid = make_kgid(net->user_ns, sysuid);
        return in_egroup_p(test_gid);

}
int _isnetauditadmin(struct net *net){
        kuid_t test_uid = make_kuid(net->user_ns, audituid);
        if (uid_eq(test_uid, current_euid())){
                return 1;
        }
        return 0;
}
int _isnetauditgroupadmin(struct net *net){
        kgid_t test_gid = make_kgid(net->user_ns, audituid);
        return in_egroup_p(test_gid);
}

/* Return standard mode bits for table entry. */
static int net_ctl_permissions(struct ctl_table_header *head,
			       struct ctl_table *table)
{
	struct net *net = container_of(head->set, struct net, sysctls);
	kuid_t root_uid = make_kuid(net->user_ns, 0);
	kgid_t root_gid = make_kgid(net->user_ns, 0);

//modify by dongshi
#if 0
	/* Allow network administrator to have same access as root. */
	if (ns_capable_noaudit(net->user_ns, CAP_NET_ADMIN)) {
		int mode = (table->mode >> 6) & 7;
		return (mode << 6) | (mode << 3) | mode;
	}
#else

	kuid_t secadm_uid = make_kuid(net->user_ns, SECADM_UID);
	kuid_t auditadm_uid = make_kuid(net->user_ns, AUDITADM_UID);
	kuid_t secau_uid = make_kuid(net->user_ns, SECAU_UID);

	kgid_t secadm_gid = make_kgid(net->user_ns, SECADM_UID);
	kgid_t auditadm_gid = make_kgid(net->user_ns, AUDITADM_UID);
	kgid_t secau_gid = make_kgid(net->user_ns, SECAU_UID);

	/* Allow network administrator to have same access as root. */
	if (ns_capable(net->user_ns, CAP_NET_ADMIN) ||
		uid_eq(root_uid, current_euid()) ||
		uid_eq(secadm_uid, current_euid()) ||
		uid_eq(secau_uid, current_euid()) ||
		_isnetadmin(net) ||
		_isnetsysadmin(net) ||
		_isnetauditadmin(net) ||
		uid_eq(auditadm_uid, current_euid())) {
		         int mode = (table->mode >> 6) & 7;
			 return (mode << 6) | (mode << 3) | mode;
	}
	/* Allow netns root group to have the same access as the root group */
	if (in_egroup_p(root_gid) || in_egroup_p(secadm_gid) || in_egroup_p(secau_gid) || in_egroup_p(auditadm_gid)||_isnetgroupadmin(net)||_isnetsysgroupadmin(net)||_isnetauditgroupadmin(net)) {
	        int mode = (table->mode >> 6) & 7;
		return (mode << 6) | (mode << 3) | mode;
	}
#endif

	return table->mode;
}

static void net_ctl_set_ownership(struct ctl_table_header *head,
				  struct ctl_table *table,
				  kuid_t *uid, kgid_t *gid)
{
	struct net *net = container_of(head->set, struct net, sysctls);
	kuid_t ns_root_uid;
	kgid_t ns_root_gid;

	ns_root_uid = make_kuid(net->user_ns, 0);
	if (uid_valid(ns_root_uid))
		*uid = ns_root_uid;

	ns_root_gid = make_kgid(net->user_ns, 0);
	if (gid_valid(ns_root_gid))
		*gid = ns_root_gid;
}

static struct ctl_table_root net_sysctl_root = {
	.lookup = net_ctl_header_lookup,
	.permissions = net_ctl_permissions,
	.set_ownership = net_ctl_set_ownership,
};

static int __net_init sysctl_net_init(struct net *net)
{
	setup_sysctl_set(&net->sysctls, &net_sysctl_root, is_seen);
	return 0;
}

static void __net_exit sysctl_net_exit(struct net *net)
{
	retire_sysctl_set(&net->sysctls);
}

static struct pernet_operations sysctl_pernet_ops = {
	.init = sysctl_net_init,
	.exit = sysctl_net_exit,
};

static struct ctl_table_header *net_header;
__init int net_sysctl_init(void)
{
	static struct ctl_table empty[1];
	int ret = -ENOMEM;
	/* Avoid limitations in the sysctl implementation by
	 * registering "/proc/sys/net" as an empty directory not in a
	 * network namespace.
	 */
	net_header = register_sysctl("net", empty);
	if (!net_header)
		goto out;
	ret = register_pernet_subsys(&sysctl_pernet_ops);
	if (ret)
		goto out1;
out:
	return ret;
out1:
	unregister_sysctl_table(net_header);
	net_header = NULL;
	goto out;
}

struct ctl_table_header *register_net_sysctl(struct net *net,
	const char *path, struct ctl_table *table)
{
	return __register_sysctl_table(&net->sysctls, path, table);
}
EXPORT_SYMBOL_GPL(register_net_sysctl);

void unregister_net_sysctl_table(struct ctl_table_header *header)
{
	unregister_sysctl_table(header);
}
EXPORT_SYMBOL_GPL(unregister_net_sysctl_table);
